/*
 * xmms-mad - mp3 plugin for xmms
 * Copyright (C) 2001-2002 Sam Clegg - See COPYING
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $superduper: decoder.c,v 1.32 2002/08/22 22:57:31 samc Exp $
 */
#include <assert.h>
#include <pthread.h>
#include <signal.h>

#include "xmms-mad.h"
#include "input.h"

#define BUFFER_SIZE 8192
/*#define DEBUG*/

/**
 * Scale PCM data
 */
static inline signed int
scale (mad_fixed_t sample)
{
  /* round */
  sample += (1L << (MAD_F_FRACBITS - 16));

  /* clip */
  if (sample >= MAD_F_ONE)
    sample = MAD_F_ONE - 1;
  else if (sample < -MAD_F_ONE)
    sample = -MAD_F_ONE;

  /* quantize */
  return sample >> (MAD_F_FRACBITS + 1 - 16);
}

void
write_output (struct mad_info_t *info, struct mad_pcm *pcm,
              struct mad_header *header)
{
  unsigned int nsamples;
  mad_fixed_t const *left_ch, *right_ch;
  char *output;
  int olen = 0;
  int pos = 0;

  nsamples = pcm->length;
  left_ch = pcm->samples[0];
  right_ch = pcm->samples[1];
  olen = nsamples * MAD_NCHANNELS (header) * 2;
  output = (char *) g_malloc (olen * sizeof (char));

  while (nsamples--)
    {
      signed int sample;
      /* output sample(s) in 16-bit signed little-endian PCM */
      sample = scale (*left_ch++);
      output[pos++] = (sample >> 0) & 0xff;
      output[pos++] = (sample >> 8) & 0xff;

      if (MAD_NCHANNELS (header) == 2)
        {
          sample = scale (*right_ch++);
          output[pos++] = (sample >> 0) & 0xff;
          output[pos++] = (sample >> 8) & 0xff;
        }
    }
  assert (pos == olen);
  mad_plugin.add_vis_pcm (mad_plugin.output->written_time (),
                          FMT_S16_LE, MAD_NCHANNELS (header), olen, output);
  while (mad_plugin.output->buffer_free () < olen)
    usleep (10000);
  mad_plugin.output->write_audio (output, olen);
  g_free (output);
}


/**
 * Decode all headers in the file and fill in stats
 */
void
scan_file (struct mad_info_t *info)
{
  struct mad_stream stream;
  struct mad_header header;
  mad_timer_t length;
  int remainder = 0;
  int len = 0;
  unsigned char buffer[BUFFER_SIZE];
  mad_stream_init (&stream);
  mad_header_init (&header);
  xing_init (&info->xing);

  length = mad_timer_zero;
  info->bitrate = 0;
  info->pos = mad_timer_zero;

  while (1)
    {
      remainder = stream.bufend - stream.this_frame;
      memcpy (buffer, stream.this_frame, remainder);
      len = input_get_data (info, buffer + remainder,
                            BUFFER_SIZE - remainder);
      if (len <= 0)
        break;

      mad_stream_buffer (&stream, buffer, len + remainder);

      while (1)
        {
          if (mad_header_decode (&header, &stream) == -1)
            {
              if (!MAD_RECOVERABLE (stream.error))
                break;
              if (stream.error == MAD_ERROR_LOSTSYNC)
                {
                  /* ignore LOSTSYNC due to ID3 tags */
                  int tagsize = id3_tag_query (stream.this_frame,
                                               stream.bufend -
                                               stream.this_frame);
                  if (tagsize > 0)
                    {
                      mad_stream_skip (&stream, tagsize);
                      continue;
                    }
                }
#ifdef DEBUG
              printf ("error decoding header %d: %s\n",
                      info->currentframe, mad_stream_errorstr (&stream));
              printf ("remainder = %d\n", remainder);
              printf ("len = %d\n", len);
#endif /* DEBUG */
              continue;
            }
          if (info->currentframe == 0)
            {
              if (!xing_parse
                  (&info->xing, stream.anc_ptr, stream.anc_bitlen))
                printf ("found xing header\n");
            }
          info->currentframe++;
          mad_timer_add (&length, header.duration);
          info->bitrate += header.bitrate;
          info->freq = header.samplerate;
        }
      if (stream.error != MAD_ERROR_BUFLEN)
        break;
    }

  mad_header_finish (&header);
  mad_stream_finish (&stream);
  xing_finish (&info->xing);

  if (info->currentframe) 
    {
      info->bitrate = info->bitrate / info->currentframe;
    }
  info->framecount = info->currentframe;
  info->currentframe = 0;
  info->length = mad_timer_count (length, MAD_UNITS_MILLISECONDS);
  /*printf ("length  = %ld\n" , length.seconds);
    printf ("frames  = %d\n" , info->framecount);
    printf ("bitrate = %d\n" , info->bitrate/1000);
    */
}

void *
decode (void *arg)
{
  char buffer[BUFFER_SIZE];
  int len;
  int seek_skip = 0;
  int remainder = 0;
  /*sigset_t signals_to_block; */

  /* mad structs */
  struct mad_stream stream;
  struct mad_frame frame;
  struct mad_synth synth;
  struct mad_info_t *info = (struct mad_info_t *) arg;

  /* sigfillset (&signals_to_block); */
  /* pthread_sigmask (SIG_BLOCK, &signals_to_block, 0); */

#ifdef DEBUG
  g_message ("f: decode");
#endif /* DEBUG */

  /* init mad stuff */
  mad_frame_init (&frame);
  mad_stream_init (&stream);
  mad_synth_init (&synth);

  if (!mad_plugin.output->open_audio (info->fmt, info->freq, info->channels))
    {
      xmmsmad_error ("failed to open audio output: %s",
                     mad_plugin.output->description);
      return 0;
    }

  /* main loop */
  do
    {
      if (info->stop)
        break;
      if (seek_skip)
        remainder = 0;
      else
        {
          remainder = stream.bufend - stream.this_frame;
          memcpy (buffer, stream.this_frame, remainder);
        }
      len = input_get_data (info, buffer + remainder,
                            BUFFER_SIZE - remainder);
      if (len <= 0)
        {
#ifdef DEBUG
          g_message ("no more data available on input");
#endif /* DEBUG */
          break;
        }
      len += remainder;
      if (len < MAD_BUFFER_GUARD)
        {
          int i;
          for (i = len; i < MAD_BUFFER_GUARD; i++)
            buffer[i] = 0;
          len = MAD_BUFFER_GUARD;
        }

      mad_stream_buffer (&stream, buffer, len);

      if (seek_skip)
        {
          int skip = 2;
          do
            {
              if (mad_frame_decode (&frame, &stream) == 0)
                {
                  mad_timer_add (&info->pos, frame.header.duration);
                  if (--skip == 0)
                    mad_synth_frame (&synth, &frame);
                }
              else if (!MAD_RECOVERABLE (stream.error))
                break;
            }
          while (skip);
          seek_skip = 0;
        }

      while (!info->stop)
        {
          if (info->seek != -1 && !info->remote)
            {
              int new_position;
              if (info->seek >= info->length)
                new_position = info->size;
              else
                new_position =
                  ((double) info->seek / (double) info->length) * info->size;
#ifdef DEBUG
              g_message ("seeking to: %d", new_position);
#endif
              mad_timer_set (&info->pos, 0, new_position, 1000);
              if (lseek (info->infile, new_position, SEEK_SET) == -1)
                xmmsmad_error ("failed to seek to position: %d",
                               new_position);
              mad_frame_mute (&frame);
              mad_synth_mute (&synth);
              stream.error = MAD_ERROR_BUFLEN;
              mad_plugin.output->flush (info->seek);
              stream.sync = 0;
              info->seek = -1;
              seek_skip = 1;
              break;
            }

          if (mad_header_decode (&frame.header, &stream) == -1)
            {
              if (!MAD_RECOVERABLE (stream.error))
                break;
              if (stream.error == MAD_ERROR_LOSTSYNC)
                {
                  /* ignore LOSTSYNC due to ID3 tags */
                  int tagsize = id3_tag_query (stream.this_frame,
                                               stream.bufend -
                                               stream.this_frame);
                  if (tagsize > 0)
                    {
                      mad_stream_skip (&stream, tagsize);
                      continue;
                    }
                }
#ifdef DEBUG
              g_message ("error decoding header %d: %s",
                      info->currentframe, mad_stream_errorstr (&stream));
#endif /* DEBUG */
              continue;
            }

          if (mad_frame_decode (&frame, &stream) == -1)
            {
              if (!MAD_RECOVERABLE (stream.error))
                break;
#ifdef DEBUG
              g_message ("error decoding frame %d: %s",
                      info->currentframe, mad_stream_errorstr (&stream));
#endif /* DEBUG */
            }

          info->currentframe++;
          if (info->currentframe % 64 == 0)
            {
              mad_plugin.set_info (info->title,
                                   (int) info->length, frame.header.bitrate,
                                   frame.header.samplerate, info->channels);
            }
          mad_synth_frame (&synth, &frame);
          mad_stream_sync (&stream);
          write_output (info, &synth.pcm, &frame.header);
          mad_timer_add (&info->pos, frame.header.duration);
        }
    }
  while (stream.error == MAD_ERROR_BUFLEN);

#ifdef DEBUG
  g_message ("end decoding");
#endif /* DEBUG */

  if (!info->stop)
    {
      mad_plugin.output->buffer_free ();
      mad_plugin.output->buffer_free ();
      while (mad_plugin.output->buffer_playing ())
        usleep (10000);
    }


  mad_plugin.output->close_audio ();

  /* reset state flags */
  info->stop = 0;
  info->playing = 0;

  pthread_exit (0);
}
